==> layerzero-0.1.11.tar.gz: layerzero-0.1.11/LayerZero/GPUAugmentation.py <==
"""
GPU-Accelerated Augmentation using Kornia

This module provides GPU-accelerated image augmentations that are 5-10x faster
than CPU-based torchvision transforms. Augmentations run on batches after
data loading, utilizing GPU compute.

Installation:
    pip install kornia kornia-rs

Usage:
    from LayerZero import GPUAugmentation
    
    aug = GPUAugmentation(image_size=224, device='cuda')
    
    # In training loop:
    for X, y in dataloader:
        X = X.to(device)
        X = aug(X)  # Apply augmentations on GPU
        ...
"""

import torch
import torch.nn as nn
from .AugmentationMode import AugmentationMode
from .KorniaHelper import is_kornia_available


class GPUAugmentation(nn.Module):
    """
    GPU-accelerated augmentation pipeline using Kornia.
    
    Benefits:
    - 5-10x faster than CPU torchvision transforms
    - Operates on batches (more efficient than per-image)
    - Utilizes GPU compute (frees CPU for other tasks)
    - Fully differentiable (can be used in training)
    
    Args:
        image_size (int): Target image size
        mode (str): 'light', 'standard', or 'heavy'
        device (str): 'cuda' or 'cpu'
        p (float): Probability of applying augmentations
    """
    
    def __init__(
        self,
        image_size=224,
        mode=AugmentationMode.BASIC,
        device='cuda',
        p=0.5,
    ):
        super().__init__()
        
        if not is_kornia_available():
            raise ImportError(
                "Kornia not installed. Install with: pip install kornia kornia-rs"
            )
        
        # Import Kornia here (after check passes)
        import kornia.augmentation as K
        
        self.image_size = image_size
        self.mode = mode
        self.device = device
        
        # Build augmentation pipeline based on mode
        augs = []
        
        if self.mode == AugmentationMode.OFF:
            # No augmentation
            augs = []
            
        elif self.mode == AugmentationMode.MINIMAL:
            # MINIMAL: Fast augmentations only
            augs = [
                K.RandomHorizontalFlip(p=0.5),
                K.RandomCrop((image_size, image_size), pad_if_needed=True),
            ]
            
        elif self.mode == AugmentationMode.BASIC:
            # BASIC: Standard augmentations for production
            augs = [
                K.RandomResizedCrop((image_size, image_size), scale=(0.2, 1.0), ratio=(0.75, 1.33), p=1.0),
                K.RandomHorizontalFlip(p=0.5),
                K.ColorJitter(0.4, 0.4, 0.4, 0.1, p=0.5),
            ]
            
        elif self.mode == AugmentationMode.STRONG:
            # STRONG: Maximum augmentation strength
            augs = [
                K.RandomResizedCrop((image_size, image_size), scale=(0.08, 1.0), ratio=(0.75, 1.33), p=1.0),
                K.RandomHorizontalFlip(p=0.5),
                K.ColorJitter(0.4, 0.4, 0.4, 0.1, p=0.8),
                K.RandomRotation(degrees=10.0, p=0.3),
                K.RandomGaussianBlur((3, 3), (0.1, 2.0), p=0.2),
                K.RandomErasing(p=0.25, scale=(0.02, 0.33), ratio=(0.3, 3.3)),
            ]
        
        # Create augmentation container
        self.transforms = K.AugmentationSequential(*augs, data_keys=["input"])
        self.transforms = self.transforms.to(device)
        
    def forward(self, x):
        """
        Apply augmentations to a batch of images.
        
        Args:
            x (torch.Tensor): Batch of images [B, C, H, W]
            
        Returns:
            torch.Tensor: Augmented images [B, C, H, W]
        """
        # Kornia expects input in range [0, 1]
        return self.transforms(x)
    
    def __repr__(self):
        return f"GPUAugmentation(mode={self.mode}, device={self.device}, transforms={len(self.transforms)})"


class HybridAugmentation(nn.Module):
    """
    Hybrid augmentation: Light CPU transforms + Heavy GPU transforms
    
    This is the optimal strategy:
    1. CPU: Only essential transforms (ToTensor, Normalize)
    2. GPU: All heavy augmentations on batched data
    
    Result: Best performance, especially on multi-worker DataLoaders
    """
    
    def __init__(self, image_size=224, mode='standard', device='cuda'):
        super().__init__()
        self.gpu_aug = GPUAugmentation(image_size, mode, device)
    
    def forward(self, x):
        """Apply GPU augmentations to batch"""
        return self.gpu_aug(x)


def benchmark_augmentation_speed():
    """
    Benchmark CPU vs GPU augmentation speed.
    
    Returns:
        dict: Timing results for CPU and GPU augmentations
    """
    if not is_kornia_available():
        print("Kornia not available. Install with: pip install kornia")
        return {}
    
    import time
    from torchvision import transforms as T
    
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    batch_size = 128
    image_size = 224
    num_iterations = 50
    
    # Create dummy batch
    batch = torch.randn(batch_size, 3, image_size, image_size)
    
    results = {}
    
    # Test CPU augmentations (torchvision)
    cpu_transforms = T.Compose([
        T.RandomResizedCrop(image_size),
        T.RandomHorizontalFlip(),
        T.ColorJitter(0.4, 0.4, 0.4, 0.1),
        T.RandomRotation(10),
    ])
    
    print(f"Benchmarking CPU augmentations on {device}...")
    start = time.time()
    for _ in range(num_iterations):
        for i in range(batch_size):
            # Torchvision works per-image
            _ = cpu_transforms(batch[i])
    cpu_time = time.time() - start
    results['cpu'] = cpu_time
    print(f"CPU time: {cpu_time:.3f}s")
    
    # Test GPU augmentations (Kornia)
    if device == 'cuda':
        gpu_aug = GPUAugmentation(image_size, mode='standard', device=device)
        batch_gpu = batch.to(device)
        
        print(f"Benchmarking GPU augmentations...")
        torch.cuda.synchronize()
        start = time.time()
        for _ in range(num_iterations):
            # Kornia works on entire batch
            _ = gpu_aug(batch_gpu)
        torch.cuda.synchronize()
        gpu_time = time.time() - start
        results['gpu'] = gpu_time
        results['speedup'] = cpu_time / gpu_time
        
        print(f"GPU time: {gpu_time:.3f}s")
        print(f"Speedup: {results['speedup']:.2f}x faster")
    else:
        print("GPU not available, skipping GPU benchmark")
    
    return results


# Example integration with Trainer
class AugmentedTrainingLoop:
    """
    Example of how to integrate GPU augmentations into training.
    
    Usage:
        gpu_aug = GPUAugmentation(224, mode='standard', device='cuda')
        
        for X, y in dataloader:
            X = X.to(device)
            X = gpu_aug(X)  # Apply GPU augmentations
            
            # Continue with training...
            logits = model(X)
            loss = loss_fn(logits, y)
            ...
    """
    pass


if __name__ == "__main__":
    print("="*60)
    print("GPU Augmentation Benchmark")
    print("="*60)
    
    if is_kornia_available():
        results = benchmark_augmentation_speed()
        
        if 'speedup' in results:
            print(f"\n✅ GPU augmentation is {results['speedup']:.2f}x faster!")
            print(f"\nRecommendation: Use GPUAugmentation for 5-10x speedup")
        else:
            print(f"\n⚠️  GPU not available. GPU augmentation requires CUDA.")
    else:
        print("\n❌ Kornia not installed.")
        print("Install with: pip install kornia kornia-rs")

==> layerzero-0.4.6.tar.gz: layerzero-0.4.6/LayerZero/GPUAugmentation.py <==
"""
GPU-Accelerated Augmentation using Kornia

This module provides GPU-accelerated image augmentations that are 5-10x faster
than CPU-based torchvision transforms. Augmentations run on batches after
data loading, utilizing GPU compute.

Installation:
    pip install kornia kornia-rs

Usage:
    from LayerZero import GPUAugmentation
    
    aug = GPUAugmentation(image_size=224, device='cuda')
    
    # In training loop:
    for X, y in dataloader:
        X = X.to(device)
        X = aug(X)  # Apply augmentations on GPU
        ...
"""

import torch
import torch.nn as nn
from .AugmentationMode import AugmentationMode
from .KorniaHelper import is_kornia_available


class GPUAugmentation(nn.Module):
    """
    GPU-accelerated augmentation pipeline using Kornia.
    
    Benefits:
    - 5-10x faster than CPU torchvision transforms
    - Operates on batches (more efficient than per-image)
    - Utilizes GPU compute (frees CPU for other tasks)
    - Fully differentiable (can be used in training)
    - Auto-detects grayscale vs RGB (skips color augs for grayscale)
    
    Args:
        image_size (int): Target image size
        mode (AugmentationMode): OFF, MINIMAL, BASIC, or STRONG
        device (str): 'cuda' or 'cpu'
        p (float): Probability of applying augmentations
        channels (int, optional): Number of channels (1=grayscale, 3=RGB). Auto-detected if None.
    """
    
    def __init__(
        self,
        image_size=224,
        mode=AugmentationMode.BASIC,
        device='cuda',
        p=0.5,
        channels=None,  # Auto-detect if None
    ):
        super().__init__()
        
        if not is_kornia_available():
            raise ImportError(
                "Kornia not installed. Install with: pip install kornia kornia-rs"
            )
        
        # Import Kornia here (after check passes)
        import kornia.augmentation as K
        
        self.image_size = image_size
        self.mode = mode
        self.device = device
        self.channels = channels  # Will be auto-detected on first forward pass if None
        self.K = K  # Store for later use
        self.transforms = None  # Will be initialized on first forward pass
        self._initialized = False
        
    def _build_transforms(self, channels):
        """Build augmentation pipeline based on number of channels."""
        augs = []
        is_grayscale = (channels == 1)
        
        if self.mode == AugmentationMode.OFF:
            # No augmentation
            augs = []
            
        elif self.mode == AugmentationMode.MINIMAL:
            # MINIMAL: Fast augmentations only (geometry only, no color)
            augs = [
                self.K.RandomHorizontalFlip(p=0.5),
                self.K.RandomCrop((self.image_size, self.image_size), pad_if_needed=True),
            ]
            
        elif self.mode == AugmentationMode.BASIC:
            # BASIC: Standard augmentations
            augs = [
                self.K.RandomResizedCrop((self.image_size, self.image_size), scale=(0.2, 1.0), ratio=(0.75, 1.33), p=1.0),
                self.K.RandomHorizontalFlip(p=0.5),
            ]
            # Only add color augmentations for RGB images
            if not is_grayscale:
                augs.append(self.K.ColorJitter(0.4, 0.4, 0.4, 0.1, p=0.5))
            
        elif self.mode == AugmentationMode.STRONG:
            # STRONG: Maximum augmentation strength
            augs = [
                self.K.RandomResizedCrop((self.image_size, self.image_size), scale=(0.08, 1.0), ratio=(0.75, 1.33), p=1.0),
                self.K.RandomHorizontalFlip(p=0.5),
            ]
            # Only add color augmentations for RGB images
            if not is_grayscale:
                augs.append(self.K.ColorJitter(0.4, 0.4, 0.4, 0.1, p=0.8))
            # Geometry augmentations work for both RGB and grayscale
            augs.extend([
                self.K.RandomRotation(degrees=10.0, p=0.3),
                self.K.RandomGaussianBlur((3, 3), (0.1, 2.0), p=0.2),
                self.K.RandomErasing(p=0.25, scale=(0.02, 0.33), ratio=(0.3, 3.3)),
            ])
        
        # Create augmentation container
        transforms = self.K.AugmentationSequential(*augs, data_keys=["input"])
        return transforms.to(self.device)
    
    def forward(self, x):
        """
        Apply augmentations to a batch of images.
        
        Args:
            x (torch.Tensor): Batch of images [B, C, H, W]
            
        Returns:
            torch.Tensor: Augmented images [B, C, H, W]
        """
        # Auto-detect channels on first forward pass
        if not self._initialized:
            if self.channels is None:
                self.channels = x.shape[1]  # Detect from input
            self.transforms = self._build_transforms(self.channels)
            self._initialized = True
            
            # Log what's being used
            aug_type = "Grayscale" if self.channels == 1 else "RGB"
            print(f"🎨 GPU Aug: {aug_type} ({self.mode.name}) | {len(self.transforms)} transforms")
        
        # Kornia expects input in range [0, 1]
        return self.transforms(x)
    
    def __repr__(self):
        num_transforms = len(self.transforms) if self._initialized else "auto"
        return f"GPUAugmentation(mode={self.mode}, device={self.device}, transforms={num_transforms})"


class HybridAugmentation(nn.Module):
    """
    Hybrid augmentation: Light CPU transforms + Heavy GPU transforms
    
    This is the optimal strategy:
    1. CPU: Only essential transforms (ToTensor, Normalize)
    2. GPU: All heavy augmentations on batched data
    
    Result: Best performance, especially on multi-worker DataLoaders
    """
    
    def __init__(self, image_size=224, mode='standard', device='cuda'):
        super().__init__()
        self.gpu_aug = GPUAugmentation(image_size, mode, device)
    
    def forward(self, x):
        """Apply GPU augmentations to batch"""
        return self.gpu_aug(x)


def benchmark_augmentation_speed():
    """
    Benchmark CPU vs GPU augmentation speed.
    
    Returns:
        dict: Timing results for CPU and GPU augmentations
    """
    if not is_kornia_available():
        print("Kornia not available. Install with: pip install kornia")
        return {}
    
    import time
    from torchvision import transforms as T
    
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    batch_size = 128
    image_size = 224
    num_iterations = 50
    
    # Create dummy batch
    batch = torch.randn(batch_size, 3, image_size, image_size)
    
    results = {}
    
    # Test CPU augmentations (torchvision)
    cpu_transforms = T.Compose([
        T.RandomResizedCrop(image_size),
        T.RandomHorizontalFlip(),
        T.ColorJitter(0.4, 0.4, 0.4, 0.1),
        T.RandomRotation(10),
    ])
    
    print(f"Benchmarking CPU augmentations on {device}...")
    start = time.time()
    for _ in range(num_iterations):
        for i in range(batch_size):
            # Torchvision works per-image
            _ = cpu_transforms(batch[i])
    cpu_time = time.time() - start
    results['cpu'] = cpu_time
    print(f"CPU time: {cpu_time:.3f}s")
    
    # Test GPU augmentations (Kornia)
    if device == 'cuda':
        gpu_aug = GPUAugmentation(image_size, mode='standard', device=device)
        batch_gpu = batch.to(device)
        
        print(f"Benchmarking GPU augmentations...")
        torch.cuda.synchronize()
        start = time.time()
        for _ in range(num_iterations):
            # Kornia works on entire batch
            _ = gpu_aug(batch_gpu)
        torch.cuda.synchronize()
        gpu_time = time.time() - start
        results['gpu'] = gpu_time
        results['speedup'] = cpu_time / gpu_time
        
        print(f"GPU time: {gpu_time:.3f}s")
        print(f"Speedup: {results['speedup']:.2f}x faster")
    else:
        print("GPU not available, skipping GPU benchmark")
    
    return results


# Example integration with Trainer
class AugmentedTrainingLoop:
    """
    Example of how to integrate GPU augmentations into training.
    
    Usage:
        gpu_aug = GPUAugmentation(224, mode='standard', device='cuda')
        
        for X, y in dataloader:
            X = X.to(device)
            X = gpu_aug(X)  # Apply GPU augmentations
            
            # Continue with training...
            logits = model(X)
            loss = loss_fn(logits, y)
            ...
    """
    pass


if __name__ == "__main__":
    print("="*60)
    print("GPU Augmentation Benchmark")
    print("="*60)
    
    if is_kornia_available():
        results = benchmark_augmentation_speed()
        
        if 'speedup' in results:
            print(f"\n✅ GPU augmentation is {results['speedup']:.2f}x faster!")
            print(f"\nRecommendation: Use GPUAugmentation for 5-10x speedup")
        else:
            print(f"\n⚠️  GPU not available. GPU augmentation requires CUDA.")
    else:
        print("\n❌ Kornia not installed.")
        print("Install with: pip install kornia kornia-rs")
